#include <stdio.h>
#include <mpi.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/time.h>
#include <algorithm>

//----------------------Rosetta includes--------------------------------------


#include<protocols/init.hh> // Rosetta initialization
#include<core/fragment/FragmentIO.hh>
#include<core/fragment/FragSet.hh>
#include <core/conformation/Conformation.hh>
//-----------------------------------------------------------------------------


#include <eo>
#include <oldmo>
#include <string>
#include "PstructFunctions.h"
#include "proteinPerturbation.h"
#include "Pstruct.h"
#include "TsFragmentMover.h"
#include "globalHCExplorer.h"
#include "proteinSmallMover.h"
#include "proteinTS.h"
#include "proteinILS.h"
#include "conformationalSample.h"
#include "moveOperators.h"
#include "singleResidueHCExplorer.h"
#include "singleFragmentResidueHCExplorer.h"
#include "simpleCoolingSchedule.h"
#include "proteinExplorerSA.h"
#include "proteinSA.h"
#include <algo/moILS.h>




using namespace std;




#include <ga.h>
#include <ga/eoBitOp.h>

#include "parserStruct.h"
#include "utils.h"
#include "load_keys.h"
#include "keyStruct.h"

#include <fstream>

using namespace core;

string pdbFile;
string scoreFunc;
core::fragment::FragmentIO fragIO;
core::fragment::FragSetOP fragDB9;
core::fragment::FragSetOP fragDB3;


Pstruct sample_minimize(parameters param,
                     vector< vector <keyStruct> > & keys,
                     vector<int> & newKeys)
{
  TsFragmentMover mover(fragDB9, param.nbFrag, "tsFragmentMover");

  Pstruct  pstruct;
  
  pstruct.create(pdbFile,scoreFunc);

  PstructEval pstruct_eval;
  
  proteinMoveInit<TsFragmentMover> move_init;
  
  proteinRandWindowNext<TsFragmentMover> window_next(mover,mover.getWindowNb());

  proteinRandMoveNext<TsFragmentMover> move_next(mover);

  proteinSingleResidueMoveNext<TsFragmentMover> single_move_next(mover, keys, param.singleSearchRate);

  proteinSingleResidueWindowNext single_window_next(pstruct.myPose.total_residue());
    
  proteinSingleRedrawWindowNext single_redraw_window_next(pstruct.myPose.total_residue());

  proteinSingleResidueRedrawMoveNext<TsFragmentMover> single_redraw_move_next(mover,keys);

  proteinMoveBestImprSelect<MoveKey, Pstruct> move_select;
 
  moGenSolContinue <Pstruct> continuBTS(param.bigtabuMaxIter);

  moGenSolContinue <Pstruct> continuSA_early(param.nbGenSA/2);

  moGenSolContinue <Pstruct> continuSA_late(param.nbGenSA);

  moGenSolContinue <Pstruct> continu3frag(param.frag3MaxIter);

  moGenSolContinue <Pstruct> continuILS(param.ILSmaxIter);

  moGenSolContinue <Pstruct> localcontinuILS(param.singleSearchMaxIter);

  proteinPerturbation<TsFragmentMover, Pstruct, MoveKey> perturbation(mover, param.perturbDegree);
  
  proteinSmallMover bbMover;
  
  moSteadyFitSolContinue <Pstruct> continuCS(param.minGenCS, param.steadyGenCS);

  conformationalSample<Pstruct, proteinSmallMover> confSampling(bbMover, continuCS, pstruct_eval); 

  singleFragmentResidueHCExplorer<TsFragmentMover, Pstruct, MoveKey> single_search_explorer(mover, move_init, single_window_next, single_move_next);

  proteinTS<TsFragmentMover, Pstruct> singleSearch(single_search_explorer, localcontinuILS, pstruct_eval);

  globalHCExplorer<TsFragmentMover, Pstruct, MoveKey> explorerTS(mover, move_init, window_next, move_next);

  reheatCoolingSchedule<Pstruct> coolingSchedule(param.initTSA, param.finalTSA, param.nbGenSA);
  
  
  proteinExplorerSA<TsFragmentMover, Pstruct, MoveKey> explorerSingleSA(mover, move_init, single_redraw_window_next, single_redraw_move_next, coolingSchedule);

                                                                      
  proteinSA<TsFragmentMover, Pstruct, MoveKey> simulatedAnnealing_singleEarly(explorerSingleSA, continuSA_early, pstruct_eval);


  proteinSA<TsFragmentMover, Pstruct, MoveKey> simulatedAnnealing_singleLate(explorerSingleSA, continuSA_late, pstruct_eval);


  proteinILS<TsFragmentMover, Pstruct, MoveKey> iterated_local_search (singleSearch, perturbation, pstruct_eval, continuILS) ;
  
  proteinTS<TsFragmentMover, Pstruct> frag3Search(explorerTS, continu3frag, pstruct_eval);

  proteinILS<TsFragmentMover, Pstruct, MoveKey> iterated_local_search3Frag (frag3Search, perturbation, pstruct_eval, continuILS) ;

  struct timeval *tv = (timeval *)malloc(sizeof(timeval));
  gettimeofday(tv,NULL);
  rng.reseed(tv->tv_usec);
  
  mover.initPose(pstruct);
  pstruct.change_score("scoring/weights/score0.wts");
  pstruct_eval(pstruct);
  simulatedAnnealing_singleEarly(pstruct);
  pstruct.change_score("scoring/weights/score1.wts");
  pstruct.invalidate();
  pstruct_eval(pstruct);
  simulatedAnnealing_singleLate(pstruct);
  pstruct.change_score("scoring/weights/score2.wts");
  pstruct.invalidate();
  pstruct_eval(pstruct);
  simulatedAnnealing_singleLate(pstruct);
  iterated_local_search(pstruct);
  pstruct.change_score("scoring/weights/score3.wts");
  pstruct.invalidate();
  pstruct_eval(pstruct);
  simulatedAnnealing_singleLate(pstruct);
  simulatedAnnealing_singleLate(pstruct);
  iterated_local_search(pstruct);
  simulatedAnnealing_singleLate(pstruct);
  simulatedAnnealing_singleLate(pstruct);
  iterated_local_search(pstruct);
  newKeys = pstruct.getAngleKeys();
  confSampling(pstruct);
  mover.set_fragments(fragDB3);
  pstruct.invalidate();
  pstruct_eval(pstruct);
  simulatedAnnealing_singleLate(pstruct);
  iterated_local_search3Frag(pstruct);
  confSampling(pstruct);
  return pstruct;
 
}


void write_scores(vector<string> files, vector<double> scores, char* scoreFile)
{
  ofstream _stream(scoreFile);
  for(unsigned i=0;i<scores.size();i++)
    {
      _stream << files[i] << " " << scores[i] << endl;
    }
  _stream.close();

}



int main( int argc, char **argv )
{
  int rank, size;
  MPI_Init( &argc, &argv );
  MPI_Comm_rank( MPI_COMM_WORLD, &rank );
  MPI_Comm_size( MPI_COMM_WORLD, &size );
    eoParser parser(argc, argv);
  parameters param;
  struct timeval *tv = (timeval *)malloc(sizeof(timeval));
  parseFile(parser, param);

  string s (argv[1]);

  if ( (s.compare("-h") == 0) || (s.compare("--help") == 0 ) )
    ;//make help
  
  pdbFile=param.inst.c_str();
  fragDB9 = fragIO.read_data(param.db9.c_str());
  fragDB3 = fragIO.read_data(param.db3.c_str());
  scoreFunc = param.scoreFunc.c_str();

  protocols::init(1, argv);


  gettimeofday(tv,NULL);
  
  TsFragmentMover mover(fragDB9, param.nbFrag, "tsFragmentMover");

  TsFragmentMover moverSA(fragDB9, 200, "SaFragmentMover");

  vector< vector <keyStruct> > keys;
  Pstruct tmpstruct;
  tmpstruct.create(pdbFile, scoreFunc);
  vector< Pstruct > pstruct;
  unsigned nbThreads = param.nbThreads;
  vector<double> fitness_tab;
  double fitness;
  int send_keys_ok;
  vector <double> scores;
  vector <string> decoyFileNames;
  unsigned participants;
  MPI_Status status;
  vector< int > value;
  vector<int> mpiKeys;
  vector< vector < double > > probas;
  vector< vector < double > > new_probas;
  mpi_keys mpiProbas;
  unsigned probas_size;
  unsigned mpiProbas_size;
  vector< vector<int> > newKeys;
  string pdbName = param.outputPdb.c_str();
  double topRate = param.topPopRate;
  unsigned nbTotIter = param.nbTotIter-1;

  rng.reseed(tv->tv_usec+rank);
  participants = size;
  fitness_tab.resize(nbThreads);
  newKeys.resize(nbThreads);
  value.resize(tmpstruct.myPose.total_residue());
  init_probas(probas, tmpstruct.myPose.total_residue(), mover.getMaxFragments(),
              mover.getWindowLength());
    if (rank == 0 )
      {
        for (unsigned nbIter=0;nbIter<nbTotIter;nbIter++)
          {
            vector< vector < int > > distribKeys;
            vector< double > energies;
            vector< vector < double > > stored_energy;
            stored_energy.resize(participants-1);
            distribKeys.resize(tmpstruct.myPose.total_residue());
            for (unsigned j=1; j<participants; j++ )
              {
                for (unsigned thread_by_node=0; thread_by_node<nbThreads; thread_by_node++)
                  {
                    
                    MPI_Recv(&fitness, 1, MPI_DOUBLE, j, 0, MPI_COMM_WORLD, &status);
                    energies.push_back(fitness);
                    stored_energy[j-1].push_back(fitness);
                  }
                sort(energies.begin(), energies.end());
              }
            double threshold_energy = energies[(int) ((participants-1)*nbThreads*topRate-1)];
            for (unsigned j=1; j<participants; j++ )
              {
                
                for (unsigned thread_by_node=0; thread_by_node<nbThreads; thread_by_node++)
                  {
               
                    if (stored_energy[j-1][thread_by_node]<=threshold_energy)
                      {
                        send_keys_ok=1;
                        MPI_Send(&send_keys_ok, 1, MPI_INT, j, 0, MPI_COMM_WORLD);
                        MPI_Recv(&value.front(), value.size(), MPI_INT, j, 0, MPI_COMM_WORLD, &status);
                        for(unsigned i=0;i<value.size();i++)
                          {
                            distribKeys[i].push_back(value[i]);
                          }
                      }
                    else
                      {
                        send_keys_ok=0;
                        MPI_Send(&send_keys_ok, 1, MPI_INT, j, 0, MPI_COMM_WORLD);
                      }
                  }
              }
            new_probas = compute_probas(distribKeys, mover.getWindowLength(), mover.getMaxFragments());
            update_probas(probas, new_probas, 0.6);
            probas_size = probas.size();
            for (unsigned j=1;j<participants;j++)
              {
                MPI_Send(&probas_size, 1, MPI_INT, j, 0, MPI_COMM_WORLD);
              }
            for (unsigned i=0;i<probas.size();i++)
              {
                mpiProbas = convert_probas_to_mpi(probas[i]);
                mpiProbas_size = mpiProbas.values.size(); 
                for (unsigned j=1;j<participants;j++)
                  {
                    MPI_Send(&mpiProbas_size, 
                             1, 
                             MPI_INT, 
                             j, 
                             0, 
                             MPI_COMM_WORLD);
                    MPI_Send(&mpiProbas.values.front(), 
                             mpiProbas.values.size(), 
                             MPI_DOUBLE, 
                             j, 
                             0, 
                             MPI_COMM_WORLD);
                    MPI_Send(&mpiProbas.indexes.front(), 
                             mpiProbas.indexes.size(), 
                             MPI_INT, 
                             j, 
                             0, 
                             MPI_COMM_WORLD);
                  }
              }
          }
      }
    else
      {
        for (unsigned thread_by_node=0; thread_by_node<nbThreads; thread_by_node++)
          {
            keys = convert_probas_to_keyStruct(probas);
            pstruct.push_back(sample_minimize(param, keys, newKeys[thread_by_node]));
          }
        for(unsigned nbIter=0; nbIter<nbTotIter; nbIter++)
          {
            for (unsigned thread_by_node=0; thread_by_node<nbThreads; thread_by_node++)
              {
                string tmpPdbName = pdbName;
                char tmpstr[20];
                sprintf(tmpstr,"_%d_process%d.pdb",nbIter+1,(rank-1)*nbThreads+thread_by_node+1);
                tmpPdbName.append(tmpstr);
                pstruct[thread_by_node].myPose.dump_pdb(tmpPdbName);
                fitness_tab[thread_by_node]=pstruct[thread_by_node].fitness();
                scores.push_back(fitness_tab[thread_by_node]);
                decoyFileNames.push_back(tmpPdbName);
                MPI_Send(&fitness_tab[thread_by_node], 1, MPI_DOUBLE, 0, 0, MPI_COMM_WORLD);
              }
            for (unsigned thread_by_node=0; thread_by_node<nbThreads; thread_by_node++)
              {
                MPI_Recv(&send_keys_ok, 1, MPI_INT, 0, 0, MPI_COMM_WORLD, &status);
                if (send_keys_ok)
                  {
                    mpiKeys = prepare_keys_mpi(newKeys[thread_by_node]);
                    MPI_Send(&mpiKeys.front(), mpiKeys.size(), MPI_INT, 0, 0, MPI_COMM_WORLD);
                  }
              }
            MPI_Recv(&probas_size, 1, MPI_INT, 0, 0, MPI_COMM_WORLD, &status);
            vector< mpi_keys > received_probas;
            for (unsigned i=0;i< probas_size; i++)
              {
                MPI_Recv(&mpiProbas_size, 1, MPI_INT, 0, 0, MPI_COMM_WORLD, &status);
                mpiProbas.values.resize(mpiProbas_size);
                mpiProbas.indexes.resize(mpiProbas_size);
                MPI_Recv(&mpiProbas.values.front(),
                         mpiProbas_size, MPI_DOUBLE,
                         0, 
                         0, 
                         MPI_COMM_WORLD, 
                         &status);
                MPI_Recv(&mpiProbas.indexes.front(), 
                         mpiProbas_size, 
                         MPI_INT, 
                         0, 
                         0, 
                         MPI_COMM_WORLD, 
                         &status);
                received_probas.push_back(mpiProbas);
              }
            keys = mpi_to_keys(received_probas);

            for (unsigned thread_by_node=0; thread_by_node<nbThreads; thread_by_node++)
              {
                
                pstruct[thread_by_node]=sample_minimize(param,
                                                     keys,
                                                     newKeys[thread_by_node]);
              }
          }
        
        for (unsigned thread_by_node=0; thread_by_node<nbThreads; thread_by_node++)
          {
            
            string tmpPdbName = pdbName;
            char tmpstr[20];
            sprintf(tmpstr,"_%d_process%d.pdb",param.nbTotIter,(rank-1)*nbThreads+thread_by_node+1);
            tmpPdbName.append(tmpstr);
            pstruct[thread_by_node].myPose.dump_pdb(tmpPdbName);
            fitness_tab[thread_by_node]=pstruct[thread_by_node].fitness();
            scores.push_back(fitness_tab[thread_by_node]);
            decoyFileNames.push_back(tmpPdbName);
          }
        char fileName[20];
        sprintf(fileName,"scores_process%d.dat",rank);
        write_scores(decoyFileNames,scores, fileName);
      }
        
  MPI_Finalize();
  return EXIT_SUCCESS;
}


