#ifndef TSFRAGMENTMOVER_H
#define TSFRAGMENTMOVER_H

#include<protocols/abinitio/FragmentMover.hh>
#include<core/fragment/FrameList.hh>
#include<core/fragment/Frame.hh>
#include<core/fragment/BBTorsionSRFD.hh>
#include<numeric/random/random.hh>
#include<core/kinematics/MoveMap.hh>
#include<vector>
#include "keyStruct.h"
#include "roulette_wheel.h"

extern numeric::random::RandomGenerator RG;
using namespace core;

class MoveKey
{
 public:
  MoveKey()
    {
      
    }
  
  MoveKey(Size _windowRank, Size _fragmentRank):
    windowRank(_windowRank),fragmentRank(_fragmentRank)
    {
    }


    
    bool operator == (MoveKey & _moveKey)
    {
      return ((windowRank==_moveKey.getWindowRank())&&(fragmentRank==_moveKey.getFragmentRank()));
    }


    bool operator == (MoveKey _moveKey)
    {
      return ((windowRank==_moveKey.getWindowRank())&&(fragmentRank==_moveKey.getFragmentRank()));
    }

    void operator = (const MoveKey & _moveKey) 
      {
        fragmentRank = _moveKey.getFragmentRank();
        windowRank = _moveKey.getWindowRank();
      }

    Size getWindowRank() const
    {
      return windowRank;
    }

    Size getFragmentRank() const
    {
      return fragmentRank;
    }

    void setWindowRank(Size _windowRank)
    {
      windowRank = _windowRank;
    }

    void setFragmentRank(Size _fragmentRank)
    {
      fragmentRank = _fragmentRank;
    }

    void setMoveKey(MoveKey _moveKey)
    {
      fragmentRank = _moveKey.getFragmentRank();
      windowRank = _moveKey.getWindowRank();
    }

 private :
    Size windowRank;
    Size fragmentRank;
};
    


class TsFragmentMover : public moMove <Pstruct>, public protocols::abinitio::FragmentMover
{
  
 public :
  typedef Pstruct::Fitness Fitness;
  
  TsFragmentMover(core::fragment::FragSetOP _fragset, Size _nbFrag, std::string _type):FragmentMover(_fragset, _type), nbFrag(_nbFrag)
    {
      windowLength = _fragset->max_frag_length();
      minOverlap = 0;
      minFraglength = windowLength;
      windowNb = insert_map_.size();
      frameNum = 1;
      maxFrag = 200;
    }
    
    void operator () (Pstruct & _sol)
    {
      cout << " You shouldn't be here : [TsFragmentMover.h, ()(_sol)" << endl;
      cout << _sol << endl;
      exit(1);

    }

    void operator () (Pstruct & _sol, MoveKey moveKey) {
      apply(_sol.myPose, moveKey);
    }


    void operator () (Pstruct & _sol, Size pos, MoveKey moveKey) {
      apply(_sol.myPose, pos, moveKey);
      
    }
        
    void operator () (Pstruct & _sol, MoveKey _moveKey, unsigned moved_residue)
    {
      core::fragment::FrameList frames;
      int windowRank;
      
      windowRank = _moveKey.getWindowRank();
      fragset_->region( *movemap(), windowRank, windowRank,
                                              minOverlap, minFraglength, frames );
      frames[frameNum]->fragment(_moveKey.getFragmentRank()).get_residue(moved_residue+1)->apply(_sol.myPose, windowRank+moved_residue);

    }
    

    void registerAngles(Pstruct & _sol, MoveKey moveKey) 
    {
          
      int wRank = moveKey.getWindowRank(); 
      for(core::Size i=0;i<windowLength;i++)
        {
          _sol.setAngleKey(wRank+i-1,i*maxFrag+(moveKey.getFragmentRank()-1));
        }
      
    }


    
    void registerAngles(Pstruct & _sol, MoveKey moveKey, double contrib) 
    {
      int wRank = moveKey.getWindowRank(); 
      for(core::Size i=0;i<windowLength;i++)
        {
          _sol.setAngleKey(wRank+i-1,i*maxFrag+(moveKey.getFragmentRank()-1));
        }
      _sol.addToContrib(wRank-1, contrib);
      
    }


    void initPose(Pstruct & _pstruct)
    {
      MoveKey _moveKey;
      for (core::Size i =1; i<windowNb+1;i++)
        {
          _moveKey.setWindowRank(i);
          select_fragment(_moveKey);
          apply(_pstruct.myPose, _moveKey);
          registerAngles(_pstruct, _moveKey);
        }
    }
    
    
    void rouletteInitPose(Pstruct & _pstruct, vector< vector <keyStruct> > & distribution)
    {
      MoveKey _moveKey;
      core::fragment::FrameList frames;
      int windowRank;

      for(core::Size i=1;i<=_pstruct.myPose.total_residue();i++)
        {
          roulette_select_move(_moveKey, distribution[i-1]);
          windowRank = i-_moveKey.getWindowRank();
          fragset_->region( *movemap(), windowRank, windowRank,
                                              minOverlap, minFraglength, frames );
          frames[frameNum]->fragment(_moveKey.getFragmentRank()).get_residue(_moveKey.getWindowRank()+1)->apply(_pstruct.myPose, i);
        }
    }
    

          

    bool valid_ss( std::string const & new_ss ) const {
      bool valid = true;
      Size helix_len = 0;
      Size strand_len = 0;
      for ( Size i = 1; i < new_ss.size(); ++i ) {
        if ( new_ss[i-1] == 'H' ) ++helix_len;
        if ( new_ss[i-1] == 'E' ) ++strand_len;
        if ( new_ss[i-1] != new_ss[i] ) {
          if ( helix_len != 0 && helix_len < 3 ) {
            valid = false;
            break;
          }
          if ( strand_len != 0 && strand_len < 2 ) {
            valid = false;
            break;
          }
          helix_len = 0;
          strand_len = 0;
        }
      } 
      return valid;
    }
    
    
    bool roulette_select_move(MoveKey & moveKey, vector<keyStruct> & probas)
    {
      int rank = roulette_wheel(probas);
       moveKey.setWindowRank(rank/maxFrag);
       moveKey.setFragmentRank((rank%maxFrag)+1);
      return true;
    }
    


    bool select_move_key(MoveKey & moveKey)
    {
      moveKey.setWindowRank(insert_map_[ static_cast< int >( rng.uniform(windowNb))+1 ]);
      Size windowRank = moveKey.getWindowRank();
      core::fragment::FrameList frames;
      fragset_->region( *movemap(), windowRank, windowRank,
                          minOverlap, minFraglength, frames );
      
      moveKey.setFragmentRank(static_cast< int >( rng.uniform(maxFrag)+ 1));
      return true;
    }
     
    
    bool smart_select_move_key(MoveKey & moveKey, vector<double> & contribs)
    {
      moveKey.setWindowRank(roulette_wheel(contribs)+1);
      Size windowRank = moveKey.getWindowRank();
      core::fragment::FrameList frames;
      fragset_->region( *movemap(), windowRank, windowRank,
                        minOverlap, minFraglength, frames );
      
      moveKey.setFragmentRank(static_cast< int >( rng.uniform(maxFrag)+ 1));
      return true;
    }

    bool select_window(MoveKey & moveKey)
    {
        
      moveKey.setWindowRank(insert_map_[ static_cast< int >( rng.uniform(windowNb)) + 1 ]);
        return true;
    }

    bool select_fragment(MoveKey & moveKey)
    {
      int rank = static_cast< int >( rng.uniform(nbFrag)) + 1;
      moveKey.setFragmentRank(rank);
      return true;
    }

    bool select_frame(core::fragment::FrameList const & frames, Size & frameNum)
    {
      frameNum = static_cast< int >( rng.uniform(frames.size())) + 1;
      return true;
    }

    void apply(core::pose::Pose & _pose)
    {
      cout << "You shouldn't be there [TsFragmentMover, aplly(pose)]" << endl;
      cout << _pose << endl;
      exit(1);
    }

    void apply(core::pose::Pose & _pose, MoveKey moveKey)
    {
        Size windowRank = moveKey.getWindowRank();
        core::fragment::FrameList frames;
        fragset_->region( *movemap(), windowRank, windowRank,
                          minOverlap, minFraglength, frames );
       
        Size fragmentRank = moveKey.getFragmentRank();
        frames[frameNum]->apply(fragmentRank, _pose);
    }
    
    void apply(core::pose::Pose & _pose, Size pos, MoveKey moveKey)
    {
      
	if( insert_size_.size() == 0 ){
		return;
	}
        
        Size const total_insert ( insert_map_.size() );
        select_window(moveKey);
        Size windowRank = moveKey.getWindowRank();
        core::fragment::FrameList frames;
        Size frameCount = fragset_->region( *movemap(), windowRank, windowRank,
                          minOverlap, minFraglength, frames );
        std::string secstruct = _pose.secstruct();

        frames[ frameNum ]->apply_ss( *movemap_, pos, secstruct );
        bool isValidSS = valid_ss( secstruct );
        if (frames[frameNum]->apply(pos, _pose) && isValidSS)
          std::cout << "Fragment applied !" << endl;
    }

    void load_frames(core::fragment::FrameList & frames, Size startFrame)
    {
      
	if( insert_size_.size() == 0 ){
		return;
	}
        

        fragset_->region( *movemap(), startFrame, startFrame,
                                            minOverlap, minFraglength, frames );
    }
    
                     
    void apply(core::pose::Pose & _pose, core::fragment::FrameList & frames, Size fragNum)
    {
        std::string secstruct = _pose.secstruct();
        Size frameNum;
        select_frame(frames, frameNum);
        frames[ frameNum ]->apply_ss( *movemap_, fragNum, secstruct );
        bool isValidSS = valid_ss( secstruct );
        if (frames[frameNum]->apply(fragNum, _pose) && isValidSS)
          ;
         
    }


    bool apply_frames(core::pose::Pose & _pose, const core::fragment::FrameList & _frameList) const
    {
      cout << "This has no effect : [TsFragmentMover: apply_frames]" << endl;
      cout << _pose << " " << _frameList << endl;
      return true;
    }


    void set_nbFrag(Size _nbFrag)
    {
      nbFrag = _nbFrag;
    }

    Size getWindowLength() const
    {
      return windowLength;
    }

    Size getNbFragments() const
    {
      return nbFrag;
    }

    Size getWindowNb() const
    {
      return windowNb;
    }

    Size getMaxFragments() const
    {
      return maxFrag;
    }

    
 private :
    Size frameNum;
    Size windowLength;
    Size windowNb;
    Size minOverlap;
    Size minFraglength;
    Size nbFrag;
    Size maxFrag;
};    


#endif
