#include <iostream>
#include <vector>
#include <fstream>
#include "keyStruct.h"

using namespace std;

typedef struct mpi_keys
{
  vector<double> values;
  vector<int> indexes;
}mpi_keys;


vector< vector <keyStruct> > load_keys(const char * _fileIn)
{
  
  char* input=(char*)malloc(sizeof(char)*256);

  ifstream fileIn;
  ifstream data(_fileIn);
  
  double key;
  keyStruct tmp;

  vector<keyStruct> residueKeys;
  vector< vector <keyStruct> > keys;


  int cpt;


  while(data >> input)
    {
      cpt=0;
      fileIn.open(input);
      if (not fileIn.is_open())
        {
          cout << "File not found :" << input << endl;
          exit(1);
        }
      while(fileIn >> key)
        {
          if (key)
            {
              tmp.value = key;
              tmp.index = cpt;
              residueKeys.push_back(tmp);
            }
          cpt++;
        }
      fileIn.clear();
      fileIn.close();
      keys.push_back(residueKeys);
      residueKeys.clear();
    }
  
  delete[] input;
  return keys;
}


int count_keys(vector<double> keys)
{
  int number_of_keys=0;
  
  for (unsigned i=0;i<keys.size();i++)
    {
      if (keys[i])
        number_of_keys++;
    }
  
  return number_of_keys;
}



vector < vector < double > > compute_probas(vector < vector < int > > & keys, 
                                          unsigned frag_size, 
                                          unsigned nb_frags)
{
  vector < vector < double > > distrib;
  distrib.resize(keys.size());
  unsigned current_nbfrags;
  for (unsigned i=0;i<keys.size();i++)
    {
      if (i<frag_size-1)
        current_nbfrags = (i+1)*nb_frags;
      else 
        current_nbfrags = frag_size*nb_frags;
      distrib[i].resize(current_nbfrags);

      for (unsigned j=0;j<keys[i].size();j++)
        {
          distrib[i][keys[i][j]]+=(double) (1./(double) keys[i].size());
        }
    }
  return distrib;
}


void init_probas(vector < vector < double > > & current_distrib,
                 unsigned nb_keys,
                 unsigned nb_frags,
                 unsigned frag_size,
                 unsigned used_frags=25)
{
  current_distrib.resize(nb_keys);
  unsigned current_nbfrags;
  for (unsigned i=0;i<nb_keys;i++)
    {
      if (i<frag_size-1)
        current_nbfrags = (i+1)*nb_frags;
      else 
        current_nbfrags = frag_size*nb_frags;
      current_distrib[i].resize(current_nbfrags);

      for (unsigned j=0;j<current_distrib[i].size();j++)
        {
          if (j<used_frags && i<=(nb_keys-frag_size))
            {
              current_distrib[i][j]=(double) (1./(double) used_frags);
            }
          else if (i>(nb_keys-frag_size))
            {
              if (j>=(frag_size-nb_keys+i)*nb_frags && j<(frag_size-nb_keys+i)*nb_frags+used_frags)
                {
                  current_distrib[i][j]=(double) (1./(double) used_frags);
                }
              else
                current_distrib[i][j]=0;
            }
          else
            current_distrib[i][j]=0;
        }
    }
}


void average_probas(vector < vector < double > > & current_distrib, 
                    vector < vector < double > > & new_distrib, 
                    unsigned nb_iters)

{
  if (current_distrib.size())
    {
      for (unsigned i=0;i<new_distrib.size();i++)
        {
          
          for (unsigned j=0;j<new_distrib[i].size();j++)
            {
              current_distrib[i][j]=((current_distrib[i][j]*nb_iters)+new_distrib[i][j])/((double) (nb_iters+1.));
            }
        }
    }
}


void update_probas(vector < vector < double > > & current_distrib, 
                    vector < vector < double > > & new_distrib, 
                    double k)

{
  if (current_distrib.size())
    {
      for (unsigned i=0;i<new_distrib.size();i++)
        {
          
          for (unsigned j=0;j<new_distrib[i].size();j++)
            {
              current_distrib[i][j]=current_distrib[i][j]*k+new_distrib[i][j]*(1-k);
            }
        }
    }
}



vector < vector < keyStruct > > convert_probas_to_keyStruct(vector< vector< double > > probas)
{
  vector < vector < keyStruct > > keys;
  keys.resize(probas.size());
  for (unsigned i=0;i<probas.size();i++)
    {
      for (unsigned j=0;j<probas[i].size();j++)
        {
          if (probas[i][j])
            {
              keyStruct tmp;
              tmp.value = probas[i][j];
              tmp.index = j;
              keys[i].push_back(tmp);
            }
        }
    }
  return keys;
}
  

mpi_keys convert_probas_to_mpi(vector<double> & keys)
{
  mpi_keys mKeys;
  int number_of_keys = count_keys(keys);
  mKeys.values.resize(number_of_keys);
  mKeys.indexes.resize(number_of_keys);
  
  int j=0;

  for(unsigned i=0;i<keys.size();i++)
    if (keys[i]!=0)
      {
        mKeys.values[j] = keys[i];
        mKeys.indexes[j] = i;
        j++;
      }

  return mKeys;
}

vector < vector < keyStruct > > mpi_to_keys(vector < mpi_keys > & probas)
{
  vector < vector < keyStruct > > keys;
  keys.resize(probas.size());
  
  for(unsigned i=0;i<probas.size();i++)
    {
      keys[i].resize(probas[i].values.size());
      for(unsigned j=0;j<probas[i].values.size();j++)
        {
          keys[i][j].value = probas[i].values[j];
          keys[i][j].index = probas[i].indexes[j];
        }
    }

  return keys;
}

vector<int> prepare_keys_mpi(vector <int> & keys)
{
  vector <int> mpi_ready;
  mpi_ready.resize(keys.size());
  for(unsigned i=0;i<keys.size();i++)
    mpi_ready[i] = keys[i];
  return mpi_ready;
}


vector< vector <int> > load_keys(const char * _fileIn, int maxFrags)
{
  
  char* input=(char*)malloc(sizeof(char)*256);

  ifstream fileIn;
  ifstream data(_fileIn);
  
  double key;
  bool present;

  vector<int> residueKeys;
  vector< vector <int> > keys;
  keys.resize(43);

  int cpt;
  int windowRank = 0;

  while(data >> input)
    {
      cpt=0;
      fileIn.open(input);
      while(fileIn >> key)
        {
          if (key)
            {
              present = false;
              for(unsigned i=0;i<keys[(int) (windowRank-cpt/maxFrags)].size();i++)
                if (keys[windowRank-(int) (cpt/maxFrags)][i]==cpt%maxFrags)
                  {
                    present = true;
                    break;
                  }
              if (!present)
                {
                  keys[(int) (windowRank-cpt/maxFrags)].push_back(cpt%maxFrags);
                }
            }
          cpt++;
        }
      fileIn.close();
      windowRank++;
    }
  
  delete[] input;

  return keys;
}


vector<double> load_contribs(const char * _fileIn, int nbWindows)
{
  ifstream fileIn(_fileIn);
  vector<double> contribs;
  double value;

  for (int i=0;i<nbWindows;i++)
    {
      fileIn >> value;
      contribs.push_back(value);
    }
  
  return contribs;
}
