#ifndef PROTEINEXPLOSA_H
#define PROTEINEXPLOSA_H

#include "keyStruct.h"
#include <coolingSchedule/moCoolingSchedule.h>

template < class M, class P, class S>
class proteinExplorerSA
{
 public :

  proteinExplorerSA(M & _mover, 
                    proteinMoveInit < M > & _init_move,
                    windowNext & _next_window,
                    moveNext & _next_move,
                    simpleCoolingSchedule<P> & _coolingSchedule):
    mover(_mover),
    init_move(_init_move),
    next_move(_next_move),
    next_window(_next_window),
    coolingSchedule(_coolingSchedule)
      {
      }
  
  virtual ~proteinExplorerSA(){}
  
  virtual void operator() (P & _sol)
  {
    if (has_next_window && (has_next_move = next_move(current)))
      {
        _sol.fitness(incremental_evaluation(mover, current, _sol));
      }
  }
    
  virtual bool validTemp()
  {
    return coolingSchedule(temperature);
  }
  

  virtual void getWindow()
  {
    next_move.init();
    has_next_window = next_window(current);
  }

  virtual void init()
  {
    init_move(mover,current);
    next_window.init();
    has_next_window = next_window(current);
    next_move.init();
    no_improve=0;
    isAccept = true;
  } 

  virtual void initParam(P & _sol)
  {
    temperature = coolingSchedule.init(_sol);
    isAccept = true;
  }
  
  virtual void terminate(P & _sol){}
  
  virtual void moveApplied(bool b)
  { 
    if (b)
      {
        if (coolingSchedule.get_reheat())
          {
            temperature = storedT;
          }
        coolingSchedule.set_reheat(false);
      }
    if (!b)
      {
        no_improve++;
        if (no_improve > coolingSchedule.get_nospan())
          {
            if (!coolingSchedule.get_reheat())
              {
                storedT = temperature;
                coolingSchedule.set_reheat(true);
              }
            no_improve = 0;
          }
      }
    has_next_move = !b;
  }
  
  virtual bool isContinue(P & _sol)
  {
    return (has_next_move && coolingSchedule(temperature));
  }
  
  virtual bool has_another_window()
  {
    return has_next_window;
  }

  virtual bool accept(P & _sol, P & _new_sol)
  {
    double alpha=0.0;
    if (_new_sol.fitness() >= _sol.fitness())
      isAccept = true;
    else
      {
        alpha = exp((_sol.fitness()-_new_sol.fitness())/temperature);
        isAccept = (rng.uniform() < alpha);
      }
    return isAccept;
  }

  double getTemperature()
  {
    return temperature;
  }

  virtual void updateParam(P & _sol)
  {
    coolingSchedule.update(temperature);
  }

 protected :
  proteinMoveIncrEval < M, S > incremental_evaluation;
  M & mover;
  proteinMoveInit < M > & init_move;
  moveNext & next_move;
  windowNext &  next_window;
  simpleCoolingSchedule<P> & coolingSchedule;
  double temperature;
  double storedT;
  S current;
  bool isAccept;
  bool has_next_move;
  bool has_next_window;
  unsigned no_improve;
};

#endif
