#ifndef PROTEINILS_H
#define PROTEINILS_H


#include "proteinTS.h"


template <class M, class P, class S>
class proteinILS
{
  typedef typename M::EOType EOT;
  typedef typename EOT::Fitness Fitness;
  
 public :
  proteinILS(proteinTS<M,P> & _tabu_search,
             proteinPerturbation<M,P,S> & _perturbation,
             eoEvalFunc<EOT> & _full_eval,
             moSolContinue<EOT> & _continu):
    tabu_search(_tabu_search),
    perturbation(_perturbation),
    full_eval(_full_eval),
    continu(_continu)
      {
        genCount = 1;
      }

    void angleStat(EOT & _solution)
    {
      for(int i=1;i<=_solution.myPose.total_residue();i++)
        {
          cout << _solution.myPose.phi(i) << " " << _solution.myPose.psi(i) << " " <<_solution.myPose.omega(i) << endl; 
        }
    }
    
    bool operator()(EOT & _solution)
    {
      if ( _solution.invalid() )
        {
          full_eval(_solution);
        }
      
      EOT _solution_saved(_solution);
      continu.init();
      tabu_search(_solution);
      if (_solution.fitness()>_solution_saved.fitness())
        {
          _solution_saved = _solution;
        }
      else
        _solution = _solution_saved;

      
      while ( continu (_solution) )
        {
          perturbation(_solution);
          _solution.invalidate();
          full_eval(_solution);
          tabu_search(_solution);
          if (_solution.fitness()>_solution_saved.fitness())
            {
              _solution_saved = _solution;
            }
          else
            _solution = _solution_saved;
          genCount++;
        }
      return true;
    }

 private:
    unsigned genCount;
    proteinTS<M,P> & tabu_search;
    proteinPerturbation<M,P,S> & perturbation;
    eoEvalFunc<EOT> & full_eval;
    moSolContinue<EOT> & continu;
    vector<EOT> history;
};

#endif
