#ifndef _SimpleCoolingSchedule_h
#define _SimpleCoolingSchedule_h
 
 
template< class EOT >
class simpleCoolingSchedule 
{
 public:
  simpleCoolingSchedule(double _initT, double _finalT, unsigned _nbGen, unsigned _span=0) : initT(_initT), span(_span), finalT(_finalT) {
    alpha = pow((double) (finalT/initT), (double) (1/(double) _nbGen));
  }
 
    virtual ~simpleCoolingSchedule() {}

    virtual double init(EOT & _solution) {

      step = 0;
 
      return initT;
    }
 
    virtual void update(double& _temp) {
      if (step >= span) {
        _temp *= alpha;
        step = 0;
      } else
        step++;
    }
 
    virtual bool operator()(double _temp) {
      return _temp > finalT;
    }
    
    virtual void set_reheat(bool b)
    {
    }

    virtual unsigned get_nospan()=0;

    virtual bool get_reheat()=0;
    
 protected:
    double initT;
    double alpha;
    unsigned int span;
    double finalT;
    unsigned int step;
};
 
template<class EOT> 
class reheatCoolingSchedule: public simpleCoolingSchedule<EOT>
{
 public :
  reheatCoolingSchedule(double _initT, double _finalT, int _nbGen, unsigned _span=0, unsigned _nospan=20000):
    simpleCoolingSchedule<EOT>(_initT, _finalT, _nbGen),
    nospan(_nospan)
    {}
  virtual ~reheatCoolingSchedule(){}
  
  virtual double init(EOT & _solution)
  {
    this->step=0;
    reheat = false;
    return this->initT;
  }


  virtual bool get_reheat()
  {
    return reheat;
  }

  virtual void set_reheat(bool b)
  {
    reheat = b;
  }

  virtual unsigned get_nospan()
  {
    return nospan;
  }

  virtual void update(double & _temp)
  {
    
    if (reheat && this->step >= this->span)
      {
        _temp*=(2-this->alpha);
        this->step = 0;
      }
    else if (this->step >= this->span) {
      if (_temp>= this->finalT)
        _temp *= this->alpha;
      this->step = 0;
    } 
    else
      this->step++;
  }

 protected :
  bool reheat;
  unsigned nospan;

};


#endif


