#ifndef SINGLEFRAGMENTRESIDUEHCEXPLORER_H
#define SINGLEFRAGMENTRESIDUEHCEXPLORER_H



#include "TsFragmentMover.h"
#include "proteinTabuList.h"
#include "proteinAspirationCriterion.h"
#include "proteinExplorer.h"
#include "moveOperators.h"


template < class M, class P, class S>
  class singleFragmentResidueHCExplorer:public proteinExplorer < P >
{
  typedef typename M::EOType EOT;
  typedef typename M::EOType::Fitness Fitness;

 public:
    singleFragmentResidueHCExplorer(M & _mover, 
                            proteinMoveInit < M > & _init_move,
                            windowNext & _next_window,
                                    moveNext & _next_move):
      proteinExplorer < P > (),
      mover(_mover),
      init_move(_init_move),
      next_move(_next_move),
      next_window(_next_window)
        {
        }
      

    void operator() (EOT & _old_solution, EOT & _new_solution)
    {
      S move, best_move;
      Fitness old_fitness, fitness, best_move_fitness;
      unsigned moved_one;
      bool  selection_update_is_ok, has_next_window, has_next_move;
      
      if( _old_solution.invalid() )
        {
          throw std::runtime_error("[singleResidueHCExplorer.h]: The current solution has not been evaluated.");
        }
      _new_solution=(EOT)_old_solution;
      EOT best_solution = _old_solution;
      init_move(mover, move);
      next_window.init();
      has_next_window = next_window(move);
      moved_one = move.getWindowRank();
      move_selector.init(_old_solution.fitness());
      do 
        {
          next_move.init();
          old_fitness = _old_solution.fitness();
          move.setWindowRank(moved_one);
          has_next_move = next_move(move);
          selection_update_is_ok=true;
          do
            {
              fitness = incremental_evaluation(mover, move, _old_solution);
              selection_update_is_ok = move_selector.update(move, fitness);
              move.setWindowRank(moved_one);
              has_next_move = next_move(move);
              if (!selection_update_is_ok)
                _old_solution = _new_solution;
            }
          while( has_next_move && !selection_update_is_ok);
          
          if (selection_update_is_ok)
            {

              move_selector(best_move, best_move_fitness);
              
              mover(_new_solution, best_move);
              mover.registerAngles(_new_solution, best_move, old_fitness - best_move_fitness);
              _new_solution.fitness(best_move_fitness);
            }
          _old_solution = _new_solution;
            
         

          has_next_window = next_window(move);
          moved_one = move.getWindowRank();
        }
      while(has_next_window);
    }
    


 private :
    M & mover;
    proteinMoveInit < M > & init_move;
    moveNext & next_move;
    windowNext &  next_window;
    singleResidueMoveBestImprSelect < S, P > move_selector;
    singleResidueMoveIncrEval < M, S > incremental_evaluation;
};








#endif
