#ifndef _UTILS_H
#define _UTILS_H
#include <stdlib.h>
#include <iostream>
#include <fstream>

using namespace std;

void parseFile(eoParser & parser, parameters & param)
{


  param.seed = parser.createParam(unsigned(time(0)), "seed", "Random number seed", 'S').value(); 

  param.popSize = parser.createParam(unsigned(15), "popSize", "Population size",'P', "Search engine" ).value();

  param.nbFrag = parser.createParam(unsigned(25), "nbFrag", "Number of fragments",'F', "Search engine" ).value();

  param.loadName = parser.createParam(string(""), "Load","A save file to restart from",'L', "Persistence" ).value();

  param.inst = parser.createParam(string(""), "inst","A data file to read instances from",'i', "Input files" ).value();

  
  param.scoreFunc = parser.createParam(string("scoring/weights/score3.wts"), "scoreFunc","A data file to read score weights",'s', "Input files" ).value();

  param.db9 = parser.createParam(string(""), "db9","A data file to read 9 fragment database",'D', "Input files" ).value();

  param.db3 = parser.createParam(string(""), "db3","A data file to read 3 fragment database",'d', "Input files" ).value();

  param.keyFile = parser.createParam(string(""), "keyFile","A data file to read fragment keys",'K', "Input files" ).value();

  param.outputPdb = parser.createParam(string("output.pdb"), "outputPdb","Output file for model",'o', "Output files" ).value();

  param.outputAngles = parser.createParam(string("angles.dat"), "outputAngles","Output file for angles",'a', "Output files" ).value();

  param.outputKeys = parser.createParam(string("keys.dat"), "outputKeys","Output file for keys",'k', "Output files" ).value();

  param.outputFitness = parser.createParam(string("fitness.dat"), "outputFitness","Output file for energy",'f', "Output files" ).value();


  param.fitnessContrib = parser.createParam(string("fitnessContrib.dat"), "fitnessContrib","Output file for fitness contribution",'c', "Output files" ).value();

  param.contribs = parser.createParam(string("contribs.dat"), "contribs","Input file for fitness contribution",'C', "Input files" ).value();


  param.schema = parser.createParam(string(""), "schema","An xml file mapping processes",'s', "Parallelism" ).value();

  param.TSmaxIter = parser.createParam(unsigned(500), "TSmaxIter", "Maximum number of iterations",'M', "Tabu Search" ).value();

  param.ILSmaxIter = parser.createParam(unsigned(10), "ILSmaxIter", "Maximum number of ILS iterations",'I', "Iterated Local Search" ).value();

  param.bigtabuMaxIter = parser.createParam(unsigned(1), "bigtabuMaxIter", "Maximum number of big tabu search iterations",'t', "Iterated Local Search" ).value();

  param.frag3MaxIter = parser.createParam(unsigned(1), "frag3MaxIter", "Maximum number of fragments size 3 iterations",'f', "Iterated Local Search" ).value();

  param.singleSearchMaxIter = parser.createParam(unsigned(1), "singleSearchMaxIter", "Maximum number of distribution based search iterations",'S', "Iterated Local Search" ).value();

  param.singleSearchRate = parser.createParam(double(0.9), "singleSearchRate", "Distribution based search rate",'r', "Iterated Local Search" ).value();
    

  param.perturbDegree = parser.createParam(unsigned(3), "perturbDegree", "Number of perturbation iterations",'p', "Iterated Local Search" ).value();

  param.tabuListSize = parser.createParam(unsigned(10), "tabuListSize", "Tabu List Size",'l', "Tabu Search" ).value();

  param.topPopRate = parser.createParam(double(0.15), "topPopRate", "perc. of pop for distribution est.",'r', "Parallelism" ).value();

  param.minGenCS = parser.createParam(int(2000), "minGenCS", "Minimum nb of gen. for conf. sampling",'Q', "Conformational sampling" ).value();

  param.steadyGenCS = parser.createParam(int(100), "steadyGenCS", "Steady nb of gen. for conf. sampling",'q', "Conformational sampling" ).value();

  param.initTSA = parser.createParam(double(3.5), "initTSA", "Initial temp. for simulated anneal.",'A', "Simulated Annealing" ).value();

  param.finalTSA = parser.createParam(double(0.5), "finalTSA", "Final temp. for simulated anneal.",'a', "Simulated Annealing" ).value();

  param.nbGenSA = parser.createParam(int(4000), "nbGenSA", "Nb of gen. for simulated anneal.",'T', "Simulated Annealing" ).value();

  param.nbTotIter = parser.createParam(unsigned(5), "nbTotIter", "Nb of stages",'R', "Iterated Local Search" ).value();

  param.nbThreads = parser.createParam(unsigned(100), "nbThreads", "Nb of runs by process",'H', "Iterated Local Search" ).value();



  string str_status = parser.ProgramName() + ".status"; 
  string statusName = parser.createParam(str_status, "status","Status file",'S', "Persistence" ).value();

  if (parser.userNeedsHelp())
    {
      parser.printHelp(cout);
      exit(1);
    }
  if (statusName != "")
    {
      ofstream os(statusName.c_str());
      os << parser;
    }
}



#endif
